import json
import logging


OMIT_ENTRIES = ['password', 'secret']


class LoggerMiddleware(object):
    log = logging.getLogger('LoggerMiddleware')

    def __init__(self, get_response):
        self.get_response = get_response

    def get_raw_data(self, request):
        pass

    def process_view(self, request, view_func, args, kwargs):
        self.log_request(request)

    def log_request(self, request):
        if request.GET or request.method in ('GET', 'DELETE'):
            data = request.GET.copy()
        elif request.POST:
            data = request.POST.copy()
        else:
            try:
                content_type = request.content_type
                if content_type == 'application/json':
                    try:
                        data = json.loads(request.body)
                    except Exception:
                        data = request.body.decode('utf8')[:3000]
                else:
                    try:
                        data = json.loads(request.body)
                    except Exception:
                        data = request.body.decode('utf8')[:3000]
            except Exception as e:
                data = '<Error: {}>'.format(e)

        if isinstance(data, dict):
            data = data.copy()
            for f in OMIT_ENTRIES:
                if f in data:
                    data[f] = '<REPLACED>'

        dct = {'meta': str(request.META), 'data': data}
        if request.FILES:
            dct['files'] = list(request.FILES.keys())
        msg = '{} {} => {}'.format(request.method, request.get_raw_uri(), json.dumps(dct))
        self.log.debug(msg)

    def __call__(self, request):
        return self.get_response(request)


class DeprecatedLoggerMiddleware(LoggerMiddleware):
    """
    Django < 1.10 compatible middleware
    """

    def __init__(self):
        pass

    def process_request(self, request):
        self.log_request(request)



import cProfile
import gc
import subprocess
from collections import defaultdict
from contextlib import contextmanager


class Profiler(cProfile.Profile):
    """
    This profiler is not intended to be used in test or production code
    Requirements: `$ pip install pyprof2calltree`
    Usage:
    >>> with Profiler.instance().profile("profile_name"):
    ...     some_code_to_profile()
    This code will generate `profile_name_1.callgrind` file. Counter increases each time profile
     called with the same name. To view these files use kcachegrind.
    """

    _instance = None

    def __init__(self, *args, **kwargs):
        self._calls = defaultdict(lambda: 0)
        if 'builtins' not in kwargs:
            kwargs['builtins'] = False
        super().__init__(*args, **kwargs)

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def _f_name(self, f_name):
        self._calls[f_name] += 1
        return "{}_{}.prof".format(f_name, self._calls[f_name])

    @contextmanager
    def profile(self, f_name):
        gc.disable()
        self.enable()
        try:
            yield
        finally:
            self.disable()
            gc.enable()
        self.dump_stats(self._f_name(f_name))
        self.clear()

    def __del__(self):
        for f_name, counter in self._calls.items():
            for i in range(1, counter + 1):
                name = "{}_{}".format(f_name, i)
                cmd = "pyprof2calltree -i {name}.prof -o {name}.calltree".format(name=name)
                subprocess.run(cmd, shell=True, executable='/bin/bash')



"""
Monitoring tools are here
Main goal is to provide easy to use monitoring solution.
Write monitoring strings to logs, ship them to kibana, extract/aggreate to monitoring system
"""
import logging

log_mon = logging.getLogger('monitoring')


def log_mon_value(name, value=1, **kwargs):
    """
    simplest monitoring function to be aggregated with sum
    """
    message = '{} => {}'.format(name, value)
    log_mon.info({'metric_name': name, 'value': value, 'message': message, **kwargs})



import asyncio
from functools import wraps
from unittest import TestCase
from unittest.case import _Outcome


class AIOMask(type):
    '''
    Sometimes test runners check test methods with
    `inspect.iscoroutine` and async functions will return `true`
    so we need to mask these methods by wrapping them into
    callable function, but this require to call such function twice
    You can find this below in:
    `await self.ensure_future(testMethod()())`
    '''
    def wrap_func(func):
        @wraps(func)
        def ret(*args, **kwargs):
            def inner():
                return func(*args, **kwargs)
            return inner
        return ret

    def __new__(cls, name, bases, dct):
        for k, v in dct.items():
            if k.startswith('test'):
                dct[k] = cls.wrap_func(v)
        return super().__new__(cls, name, bases, dct)


class AIOTestCase(TestCase, metaclass=AIOMask):
    '''
    Basically this was just copy-paste from native python test case implementation
    with some changes to test async/await methods.
    It require AIOMask metaclass to deal with some test runners like `nose2` and etc,
    you can read about this in AIOMask docstring
    We provide async support for `setUp`, `tearDown` and test methods, also support async
    functions in `assertRaises`
    '''
    def run(self, result=None):
        self.loop = asyncio.get_event_loop()
        out = self.inner_run(result)
        f = asyncio.ensure_future(out)
        self.loop.run_until_complete(f)
        return f.result()

    async def assertRaises(self, CatchException, callable, *args, **kwargs):
        try:
            await self.ensure_future(callable(*args, **kwargs))
        except CatchException:
            pass

    def ensure_future(self, result):
        try:
            return asyncio.ensure_future(result)
        except TypeError:
            f = asyncio.Future()
            f.set_result(result)
            return f

    async def inner_run(self, result=None):
        orig_result = result
        if result is None:
            result = self.defaultTestResult()
            startTestRun = getattr(result, 'startTestRun', None)
            if startTestRun is not None:
                startTestRun()

        result.startTest(self)

        testMethod = getattr(self, self._testMethodName)
        if (getattr(self.__class__, "__unittest_skip__", False) or
           getattr(testMethod, "__unittest_skip__", False)):
            # If the class or method was skipped.
            try:
                skip_why = (getattr(self.__class__, '__unittest_skip_why__', '')
                            or getattr(testMethod, '__unittest_skip_why__', ''))
                self._addSkip(result, self, skip_why)
            finally:
                result.stopTest(self)
            return
        expecting_failure_method = getattr(testMethod,
                                           "__unittest_expecting_failure__", False)
        expecting_failure_class = getattr(self,
                                          "__unittest_expecting_failure__", False)
        expecting_failure = expecting_failure_class or expecting_failure_method
        outcome = _Outcome(result)
        try:
            self._outcome = outcome

            with outcome.testPartExecutor(self):
                await self.ensure_future(self.setUp())
            if outcome.success:
                outcome.expecting_failure = expecting_failure
                with outcome.testPartExecutor(self, isTest=True):
                    # wrapped into function to prevent generator check return true
                    await self.ensure_future(testMethod()())
                outcome.expecting_failure = False
                with outcome.testPartExecutor(self):
                    await self.ensure_future(self.tearDown())

            self.doCleanups()
            for test, reason in outcome.skipped:
                self._addSkip(result, test, reason)
            self._feedErrorsToResult(result, outcome.errors)
            if outcome.success:
                if expecting_failure:
                    if outcome.expectedFailure:
                        self._addExpectedFailure(result, outcome.expectedFailure)
                    else:
                        self._addUnexpectedSuccess(result)
                else:
                    result.addSuccess(self)
            return result
        finally:
            result.stopTest(self)
            if orig_result is None:
                stopTestRun = getattr(result, 'stopTestRun', None)
                if stopTestRun is not None:
                    stopTestRun()

            # explicitly break reference cycles:
            # outcome.errors -> frame -> outcome -> outcome.errors
            # outcome.expectedFailure -> frame -> outcome -> outcome.expectedFailure
            outcome.errors.clear()
            outcome.expectedFailure = None

            # clear the outcome, no more needed
            self._outcome = None



import itertools
import os
import sys
import zipfile
from decimal import ROUND_HALF_UP, Decimal
from pathlib import Path
from typing import Any, BinaryIO, Dict, List, Union


def execfile(fname, _globals, _locals):
    """
    Usage: execfile('path/to/file.py', globals(), locals())
    """
    if os.path.exists(fname):
        with open(fname) as f:
            code = compile(f.read(), os.path.basename(fname), 'exec')
            exec(code, _globals, _locals)
            return True
    else:
        return False


def rel_path(path, check=False, depth=1):
    d = os.path.dirname(sys._getframe(depth).f_code.co_filename)
    full = os.path.abspath(os.path.join(d, path))
    if check and not os.path.exists(full):
        raise Exception('No such path: {!r}'.format(full))
    return full


def py_rel_path(*args, **kwargs):
    return Path(rel_path(*args, depth=2, **kwargs))


def auto_directory(rel_name):
    """
    if you're using py.path you make do that as:
    py.path.local(full_path).ensure_dir()
    """
    dir_name = rel_path(rel_name, check=False)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name, exist_ok=True)
    return dir_name


def usd_round(amount):
    if isinstance(amount, str):
        amount = Decimal(amount)
    return amount.quantize(Decimal('.01'), rounding=ROUND_HALF_UP)


def root_directory(target: Union[Path, BinaryIO], exclude: List[str] = None) -> Path:
    """Returns the folder that contains the files."""
    exclude = exclude or []
    if isinstance(target, Path):
        files = [f for f in target.glob('**/*') if f.is_file()]
    elif hasattr(target, 'read'):
        with zipfile.ZipFile(target, 'r') as zip_ref:
            files = [
                Path(zip_item.filename) for zip_item in zip_ref.infolist() if not zip_item.is_dir()
            ]
    else:
        raise Exception('Unsupported target object')
    paths = [f for f in files if not any(filter(lambda i: i in str(f), exclude))]
    if len(paths) > 1:
        prefix = os.path.commonpath(paths)
        prefix = '' if prefix == '/' else prefix
        root = Path(prefix)
    elif paths:
        root = paths[0].parent
    else:
        root = Path('')
    return root


def dict_contains(superset, subset):
    """
    partial comparison dictionaries

    $ dict_contains({'a': 1, 'b': 2}, {'a': 1}) # => True
    $ dict_contains({'a': 1, 'b': 2}, {'b': 1}) # => True
    $ dict_contains({'a': 1, 'b': 2}, {'c': 3}) # => False
    $ dict_contains({'a': 1, 'b': 2}, {'b': 2}) # => False
    """

    for k, v in subset.items():
        if k not in superset:
            return False
        if isinstance(v, dict):
            if not dict_contains(superset[k], subset[k]):
                return False
        else:
            if v != superset[k]:
                return False
    return True


def slide(iterable, size=2):
    return itertools.zip_longest(*[itertools.islice(iterable, i, sys.maxsize) for i in range(size)])


def expand_dot(pattern: Dict[str, Any]):
    """
    {'a.b.c': 1} => {'a':{'b':{'c': 1}}}
    """
    out = {}
    for k, v in pattern.items():
        if '.' in k:
            curr = out
            for sub_k, has_next in slide(k.split('.')):
                if has_next:
                    curr = curr.setdefault(sub_k, {})
                else:
                    curr[sub_k] = v
        else:
            out[k] = v
    return out



import json

from unittest import mock
from rest_framework import status


class MockResponse:
    def __init__(self, json_content, status_code):
        self.json_content = json_content
        self.status_code = status_code

    def json(self, **params):
        return self.json_content

    @property
    def content(self):
        return json.dumps(self.json_content)


class MockGet:
    target = 'requests.get'

    @classmethod
    def mock(cls, result, status_code=status.HTTP_200_OK):
        """
        Patch requests.get method to always return specified response
        """
        def deco(fun):
            def new_fun(*args, **kwargs):
                def get_results(*inner_args, **inner_kwargs):
                    if callable(result):
                        self = args[0]
                        r = result(self)
                        return MockResponse(r, status_code)
                    else:
                        return MockResponse(result, status_code)

                patcher, m = cls.get_patcher(get_results)
                patcher.start()
                try:
                    ans = fun(mock_obj=m, *args, **kwargs)
                finally:
                    patcher.stop()
                return ans
            return new_fun
        return deco

    @classmethod
    def get_patcher(cls, side_effect):
        m = mock.Mock(side_effect=side_effect)
        return mock.patch(cls.target, m), m


class MockPost(MockGet):
    target = 'requests.post'


class MockPatch(MockGet):
    target = 'requests.patch'


class MockPut(MockGet):
    target = 'requests.put'


class MockDelete(MockGet):
    target = 'requests.delete'



import json
import os
from importlib import import_module

from docutils import nodes
from docutils.parsers.rst import Directive

from .parse_mixin import ParseMixin


def import_name(name):
    m_name, c_name = name.rsplit('.', 1)
    return getattr(import_module(m_name), c_name)


def p(text, *children):
    return nodes.paragraph('', text, *children)


def parse_doc(doc):
    """
    Parse docstrings to dict, it should look like:
      key: value
    """
    if not doc:
        return {}
    out = {}
    for s in doc.split('\n'):
        s = s.strip().split(':', maxsplit=1)
        if len(s) == 2:
            out[s[0]] = s[1]
    return out


class SerializerField:
    def _method_field(self, method):
        if method['method_name']:
            params = parse_doc(method['method_doc'])
            if 'type' in params:
                _type = params['type']
                return self.parse(_type) if ':any:' in _type else p(params['type'])
            return p('METHOD_NEED_TYPE_STUB: {}'.format(method['method_name']))
        return p('METHOD_FIELD_STUB')

    def _dyn_field(self, ref_name):
        return self.parse(':any:`{}`'.format(ref_name))

    def _dyn_list(self, ref_name):
        return self.parse(':any:`list[{name}] <{name}>`'.format(name=ref_name))

    @staticmethod
    def _none_field(data):
        tpl = 'FHELPER_TYPE_NOT_DESCRIBED_STUB: {} DM: {} {}'
        return p(tpl.format(data['__field__'], data['is_dyn_serializer'], data['__class__']))

    @staticmethod
    def _doc_field(doc):
        params = parse_doc(doc['model_doc'])
        if 'type' not in params:
            params = parse_doc(doc['serializer_doc'])
        return p(params['type']) if 'type' in params else None

    def __init__(self, f, directive):
        self.f = f
        self.directive = directive
        self.parse = directive.parse

    def field_type(self, field):
        field_type = None
        type_data = field['type']

        if 'primitive_type' in type_data:
            field_type = p(type_data['primitive_type'])
        elif 'method_type' in type_data:
            field_type = self._method_field(type_data['method_type'])
        elif 'dyn_field_type' in type_data:
            field_type = self._dyn_field(type_data['dyn_field_type']['ref_name'])
        elif 'list_field_type' in type_data:
            child_type = type_data['list_field_type']
            if 'ref_name' in child_type:
                field_type = self._dyn_list(type_data['list_field_type']['ref_name'])
            elif 'primitive_type' in child_type:
                p_type = child_type['primitive_type']
                field_type = p('list[{}]'.format(p_type))
            else:
                pass  # TODO: Process other cases
        elif 'doc_type' in type_data:
            field_type = self._doc_field(type_data['doc_type'])

        if not field_type:
            field_type = self._none_field(field)

        field_type += nodes.Text(' ')

        if field['readonly']:
            field_type += self.parse('|readonly|')[0]
        if field['write_only']:
            field_type += self.parse('|writeonly|')[0]
        if field['required']:
            field_type += self.parse('|required|')[0]
        if field['allow_null']:
            field_type += self.parse('|nullable|')[0]

        return field_type

    def process(self):
        field = self.f

        field_name = nodes.entry('', p(field['name']))
        field_type = nodes.entry('', self.field_type(field))
        field_description = nodes.entry('', p(nodes.Text(field.get('help_text', '') or '')))

        content = nodes.row('', field_name, field_type, field_description)
        return content


class DynSerializer(Directive, ParseMixin):
    required_arguments = 1
    has_content = True

    def run(self):
        artifact_path = os.environ.get('DOCS_ROOT', './.doc')
        full_path = '{}.json'.format(os.path.join('../', artifact_path, self.arguments[0]))
        with open(full_path) as f:
            obj_data = json.load(f)

        p_fields = (SerializerField(f, self).process() for f in obj_data['fields'])

        class_name = obj_data['class_name']

        table_fields = nodes.tbody('')
        table_fields.extend(p_fields)

        name = nodes.title('', '{}'.format(class_name))
        table_head = nodes.thead(
            '',
            nodes.row(
                '',
                nodes.entry('', p('name')),
                nodes.entry('', p('type')),
                nodes.entry('', p('description')),
            ),
        )
        tbl = nodes.table()
        tg = nodes.tgroup()
        tbl.append(tg)
        tg.append(nodes.colspec(colwidth=1))
        tg.append(nodes.colspec(colwidth=8))
        tg.append(nodes.colspec(colwidth=8))
        tg.append(table_head)
        tg.append(table_fields)

        class_descr = 'Class: *{}.{}*'.format(obj_data['class_module'], class_name)
        filter_field_param = 'Filter field: *{}*'.format(obj_data['field_param'])
        description = (obj_data.get('class_doc', '') or '').strip()
        full_name = self.parse(class_descr)
        filter_param = self.parse(filter_field_param)

        sections = ['', name]
        if description:
            sections.append(self.parse(description))
        sections.extend([full_name, filter_param, p('', nodes.topic('', tbl))])

        raw = nodes.section(*sections)

        target = self.parse('.. _{name}:'.format(name=class_name))
        return [target, raw]


def setup(app):
    app.add_directive('dyn_serializer', DynSerializer)
    return {'version': '0.0.1'}



import sys
from contextlib import suppress

from fan_tools.drf.serializers import EnumSerializer

from django.core.exceptions import ImproperlyConfigured
from django.core.validators import RegexValidator
from django.db.models import AutoField, ForeignKey, OneToOneField
from rest_framework import fields
from rest_framework.fields import ListField
from rest_framework.serializers import ListSerializer
from rest_framework_dyn_serializer import DynModelSerializer

try:
    from django.contrib.gis.db.models import MultiPointField
except ImproperlyConfigured as e:
    print(
        f'WARNING: Can not import MultiPointField. Fields with this type will be skipped. ({e})',
        file=sys.stderr,
    )
    MultiPointField = type(None)


def fields_regexp(field):
    pattern = '*'
    for validator in field.validators:
        if isinstance(validator, RegexValidator):
            pattern = validator.regex.pattern

    return f'str regex({pattern})'


def fields_enum(field):
    values = ', '.join(v.name for v in field.enum)
    return f'str enum({values})'


def fields_choices(field):
    # TODO: choices type - str vs int
    values = ', '.join(str(name) for name in field.choices)
    return f'str enum({values})'


def fields_model(field):
    return django_types.get(field.model_field.__class__)


django_types = {MultiPointField: 'list[(x:float, y:float), ..]'}


primitive_types = {
    fields.IntegerField: 'int',
    fields.BooleanField: 'bool',
    fields.FloatField: 'float',
    fields.CharField: 'str',
    fields.ImageField: 'http-link',
    fields.DateTimeField: 'datetime',
    fields.DateField: 'date',
    fields.JSONField: 'json/object',
    fields.URLField: 'http-link',
    fields.RegexField: fields_regexp,
    fields.EmailField: 'email',
    fields.ChoiceField: fields_choices,
    fields.DecimalField: 'decimal',
    fields.ModelField: fields_model,
    EnumSerializer: fields_enum,
}


def primitive_type(cls, field):
    for field_class, field_type in primitive_types.items():
        if issubclass(cls, field_class):
            return {'primitive_type': field_type(field) if callable(field_type) else field_type}

    return None


def serializer_doc_info(serializer, path_info=''):
    print('Run: {}'.format(path_info))

    if not hasattr(serializer, 'Meta'):
        serializer.Meta = type("Meta", tuple(), {})
    if not hasattr(serializer.Meta, 'fields_param'):
        serializer.Meta.fields_param = 'any'

    serializer_obj = serializer()

    return {
        'class_name': serializer.__name__,
        'class_module': serializer.__module__,
        'field_param': serializer.Meta.fields_param,
        'fields': [
            process_fields(name, field, serializer_obj)
            for name, field in sorted(serializer_obj.fields.items())
        ],
        'class_doc': serializer.__doc__,
    }


def process_fields(name, field, serializer):
    cls = field.__class__

    model = None
    model_field = None
    if hasattr(serializer, 'Meta') and hasattr(serializer.Meta, 'model'):
        model = serializer.Meta.model
        with suppress(Exception):
            model_field = model._meta.get_field(name)

    # ManyToOneRel and ManyToManyRel don't have help_text
    help_text = field.help_text or getattr(model_field, 'help_text', None)

    result = {
        'name': name,
        'readonly': bool(field.read_only),
        'write_only': bool(field.write_only),
        'required': bool(field.required),
        'allow_null': bool(field.allow_null),
        'is_dyn_serializer': issubclass(cls, DynModelSerializer),
        'help_text': help_text,
        '__field__': repr(field),
        '__class__': repr(cls),
        'type': get_type(field, serializer, model_field, model, name),
    }

    print('Process: <SerializerField: {}/>'.format(name))
    return result


def get_type(field, serializer, model_field, model, name):
    cls = field.__class__

    p_type = primitive_type(cls, field)
    if p_type:
        return p_type

    if cls == fields.SerializerMethodField:
        method_name = field.method_name or 'get_{}'.format(name)
        method = getattr(serializer, method_name)
        return {
            'method_type': {
                'method_name': method_name,
                'method_doc': method.__doc__ if method and method.__doc__ else None,
            }
        }

    if issubclass(cls, DynModelSerializer):
        return {'dyn_field_type': {'ref_name': cls.__name__}}

    if issubclass(cls, ListSerializer) and isinstance(field.child, DynModelSerializer):
        return {'list_field_type': {'ref_name': field.child.__class__.__name__}}

    if issubclass(cls, ListField):
        p_type = primitive_type(field.child.__class__, field.child)
        if p_type:
            return {'list_field_type': p_type}

    if model_field is not None and isinstance(model_field, ForeignKey):
        # if issubclass(model_field, ForeignKey):
        related_field = model_field.target_field
        while isinstance(related_field, OneToOneField):  # Table inheritance
            related_field = related_field.target_field
        related_types = {AutoField: 'int (related)'}
        if related_field.__class__ in related_types:
            return {'primitive_type': related_types[related_field.__class__]}

    model_doc = None
    if model is not None:
        # parse docstring in custom model methods
        fun = getattr(model, name, {})
        model_doc = fun.__doc__

    serializer_doc = field.__doc__
    if model_doc or serializer_doc:
        return {'doc_type': {'model_doc': model_doc, 'serializer_doc': serializer_doc}}

    return {}



import asyncio
import logging
import os
import re
import socket
import subprocess
import time
from asyncio.subprocess import PIPE
from collections import ChainMap
from contextlib import closing, contextmanager, suppress

log = logging.getLogger('fan_tools.unix')


class ExecError(Exception):
    def __init__(self, exit_code, cmd, stdout, stderr):
        self.exit_code = exit_code
        self.cmd = cmd
        self.stdout = stdout
        self.stderr = stderr
        super().__init__()

    def __str__(self):
        return f'Return: {self.exit_code} Command: {self.cmd}\nStderr: {self.stderr}'

    def __repr__(self):
        desc = str(self).replace('\n', '\t')
        return f'<ExecException `{desc}`>'


def _prepare(out):
    out = out.decode('utf8').strip('\n').split('\n')
    if out == ['']:
        return []
    return out


def run(command):
    '''
    Run command in shell, accepts command construction from list
    Return (return_code, stdout, stderr)
    stdout and stderr - as list of strings
    '''
    if isinstance(command, list):
        command = ' '.join(command)
    out = subprocess.run(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return (out.returncode, _prepare(out.stdout), _prepare(out.stderr))


def succ(cmd, check_stderr=False, stdout=None, stderr=None):
    '''
    Alias to run with check return code and stderr
    '''
    code, out, err = run(cmd)

    # Because we're raising error, sometimes we want to process stdout/stderr after catching error
    # so we're copying these outputs if required
    if stdout is not None:
        stdout[:] = out
    if stderr is not None:
        stderr[:] = err

    if code != 0:
        for l in out:
            print(l)
    if code != 0:
        raise ExecError(code, cmd, out, err)
    if check_stderr and err != []:
        raise ExecError(code, cmd, out, err)
    return code, out, err


async def process_pipe(out, pipe, proc, log_fun):
    while True:
        line = await pipe.readline()
        line = line.decode('utf8').strip('\n')
        if line:
            out.append(line)
            log_fun(line)
        if not line and proc.returncode is not None:
            return
        await asyncio.sleep(0)


async def asucc(
    cmd, check_stderr=False, pid_future=None, with_log=True, stdout=None, stderr=None, loop=None
):
    proc = await asyncio.create_subprocess_shell(
        cmd, stdin=PIPE, stderr=PIPE, stdout=PIPE, loop=loop
    )
    if pid_future and not pid_future.done():
        pid_future.set_result(proc.pid)

    # see succ comments for these values
    out = stdout if stdout is not None else []
    err = stderr if stderr is not None else []

    if with_log:
        log_warning = log.warning
        log_debug = log.debug
    else:
        skip = lambda x: x
        log_warning, log_debug = skip, skip

    asyncio.ensure_future(process_pipe(err, proc.stderr, proc, log_warning), loop=loop)
    asyncio.ensure_future(process_pipe(out, proc.stdout, proc, log_debug), loop=loop)

    try:
        code = await proc.wait()
        if code != 0:
            raise ExecError(code, cmd, out, err)
        if check_stderr and err != []:
            raise ExecError(code, cmd, out, err)
        return code, out, err
    except asyncio.CancelledError as e:
        if not proc.returncode:
            log.exception('Going to kill process: [{}] {}. Children first'.format(proc.pid, cmd))
            await asucc('pkill -9 -P {} || true'.format(proc.pid))
            with suppress(ProcessLookupError):
                proc.terminate()
            await asyncio.sleep(0.1)
            if not proc.returncode:
                with suppress(ProcessLookupError):
                    proc.kill()
        raise e


def check_socket(host, port):
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.settimeout(10)
        try:
            conn_ex = sock.connect_ex((host, port))
            if conn_ex == 0:
                return True
            return False
        except:
            return False


def wait_result(func, result, timeout):
    t = time.time()
    while True:
        res = func()
        if result is None and res is None:
            return True
        elif res == result:
            return True
        time.sleep(1)
        if (time.time() - t) > timeout:
            return False


def wait_socket(host, port, timeout=120):
    '''
    Wait for socket opened on remote side. Return False after timeout
    '''
    return wait_result(lambda: check_socket(host, port), True, timeout)


def wait_no_socket(host, port, timeout=120):
    return wait_result(lambda: check_socket(host, port), False, timeout)


def interpolate_sysenv(line, defaults={}):
    '''
    Format line system environment variables + defaults
    '''
    map = ChainMap(os.environ, defaults)
    return line.format(**map)


def source(fname):
    '''
    Acts similar to bash 'source' or '.' commands.
    '''
    rex = re.compile('(?:export |declare -x )?(.*?)="(.*?)"')
    out = call_out('source {} && export'.format(fname))
    out = [x for x in out if 'export' in x or 'declare' in x]
    out = {k: v for k, v in [rex.match(x).groups() for x in out if rex.match(x)]}
    for k, v in out.items():
        os.environ[k] = v


@contextmanager
def cd(dir_name):
    """
    do something in other directory and return back after block ended
    """
    old_path = os.path.abspath('.')
    os.chdir(dir_name)
    try:
        yield
        os.chdir(old_path)
    except Exception:
        os.chdir(old_path)
        raise



from enum import Enum, IntEnum

from rest_framework import fields


class EnumSerializer(fields.Field):
    default_error_messages = {'invalid': '{value!r} not valid value for: {values}'}

    def __init__(self, enum, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enum = enum

    def to_internal_value(self, data):
        if data is None:
            return None
        try:
            return self.enum[data]
        except (KeyError, ValueError):
            try:
                # isinstance(self.enum, IntEnum) => generates False for IntEnums
                if IntEnum in self.enum.mro() and isinstance(data, str):
                    data = int(data)
                return self.enum(data)
            except ValueError:
                values = [x.name for x in self.enum]
                self.fail('invalid', value=data, values=values)
        except AttributeError:
            values = [x.name for x in self.enum]
            self.fail('invalid', value=data, values=values)

    def to_representation(self, value):
        if isinstance(value, Enum):
            return value.name
        return self.enum(value).name



class PropsMeta(type):
    '''
    Replace all methods prop_[METHODNAME] with property
    This property return method result and cache it between calls
    Cached result stored in __[METHODNAME]
    '''
    def __new__(cls, name, bases, attrs, **kwargs):
        new_attrs = cls.new_attrs(attrs)
        return type.__new__(cls, name, bases, new_attrs)

    PROP = 'prop__'

    @classmethod
    def new_attrs(cls, attrs):
        new_attrs = {}

        def is_empty(self, name):
            return not hasattr(self, name)

        for name, value in list(attrs.items()):
            if not name.startswith(cls.PROP):
                new_attrs[name] = value
                continue
            new_name = name.replace(cls.PROP, '')
            attr_name = '__{}'.format(new_name)

            def _prop_wrapper(value_func, attr_name):
                def wrapped(self):
                    if is_empty(self, attr_name):
                        setattr(self, attr_name, value_func(self))
                    return getattr(self, attr_name)
                return wrapped

            new_attrs[new_name] = property(_prop_wrapper(value, attr_name))
        return new_attrs
